"""Classes to interact with devices."""
from datetime import datetime

from .const import (
    ROOT_DEVICES,
    ATTR_APPLICATION_TYPE,
    ATTR_DEVICE_INFO,
    ATTR_NAME,
    ATTR_CREATED_AT,
    ATTR_ID,
    ATTR_REACHABLE_STATE,
    ATTR_LAST_SEEN,
    ATTR_LIGHT_CONTROL,
    ATTR_LIGHT_STATE,
    ATTR_LIGHT_DIMMER,
    ATTR_LIGHT_COLOR_X,
    ATTR_LIGHT_COLOR_Y,
    ATTR_LIGHT_COLOR
)


class Device(object):
    """Base class for devices."""

    def __init__(self, api, raw):
        self.api = api
        self.raw = raw

    @property
    def id(self):
        return self.raw.get(ATTR_ID)

    @property
    def application_type(self):
        return self.raw.get(ATTR_APPLICATION_TYPE)

    @property
    def name(self):
        return self.raw.get(ATTR_NAME)

    @property
    def path(self):
        return [ROOT_DEVICES, self.id]

    @property
    def device_info(self):
        return DeviceInfo(self)

    @property
    def created_at(self):
        if ATTR_CREATED_AT not in self.raw:
            return None
        return datetime.utcfromtimestamp(self.raw[ATTR_CREATED_AT])

    @property
    def last_seen(self):
        if ATTR_LAST_SEEN not in self.raw:
            return None
        return datetime.utcfromtimestamp(self.raw[ATTR_LAST_SEEN])

    @property
    def reachable(self):
        return self.raw.get(ATTR_REACHABLE_STATE) == 1

    @property
    def has_light_control(self):
        return (ATTR_LIGHT_CONTROL in self.raw and
                len(self.raw.get(ATTR_LIGHT_CONTROL)) > 0)

    @property
    def light_control(self):
        return LightControl(self)

    def set_name(self, name):
        """Set group name."""
        self.set_values({
            ATTR_NAME: name
        })

    def set_values(self, values):
        """Helper to set values for device."""
        self.api('put', self.path, values)

    def update(self):
        self.raw = self.api('get', self.path)

    def __repr__(self):
        return "<{} - {} ({})>".format(self.id, self.name,
                                       self.device_info.model_number)


class DeviceInfo:
    """Represent device information.

    http://www.openmobilealliance.org/tech/profiles/LWM2M_Device-v1_0.xml
    """
    ATTR_MANUFACTURER = "0"
    ATTR_MODEL_NUMBER = "1"
    ATTR_SERIAL = "2"
    ATTR_FIRMWARE_VERSION = "3"
    ATTR_POWER_SOURCE = "6"
    VALUE_POWER_SOURCES = {
        1: 'Internal Battery',
        2: 'External Battery',
        3: 'Battery',  # Not in spec, used by remote
        4: 'Power over Ethernet',
        5: 'USB',
        6: 'AC (Mains) power',
        7: 'Solar'
    }
    ATTR_BATTERY = "9"

    def __init__(self, device):
        self._device = device

    @property
    def manufacturer(self):
        """Human readable manufacturer name."""
        return self.raw.get(DeviceInfo.ATTR_MANUFACTURER)

    @property
    def model_number(self):
        """A model identifier string (manufactuer specified string)."""
        return self.raw.get(DeviceInfo.ATTR_MODEL_NUMBER)

    @property
    def serial(self):
        """Serial number string."""
        return self.raw.get(DeviceInfo.ATTR_SERIAL)

    @property
    def firmware_version(self):
        """Current firmware version string of the device."""
        return self.raw.get(DeviceInfo.ATTR_FIRMWARE_VERSION)

    @property
    def power_source(self):
        """Power source."""
        return self.raw.get(DeviceInfo.ATTR_POWER_SOURCE)

    @property
    def power_source_str(self):
        """String representation of current power source."""
        if DeviceInfo.ATTR_POWER_SOURCE not in self.raw:
            return None
        return DeviceInfo.VALUE_POWER_SOURCES.get(self.power_source, 'Unknown')

    @property
    def battery_level(self):
        """Battery in 0..100"""
        return self.raw.get(DeviceInfo.ATTR_BATTERY)

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self._device.raw[ATTR_DEVICE_INFO]


class LightControl:
    """Class to control the lights."""

    def __init__(self, device):
        self._device = device

    @property
    def lights(self):
        """Return light objects of the light control."""
        return [Light(self._device, i) for i in range(len(self.raw))]

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self._device.raw[ATTR_LIGHT_CONTROL]

    def set_state(self, state, *, index=0):
        """Set state of a light."""
        self.set_values({
            ATTR_LIGHT_STATE: int(state)
        }, index=index)

    def set_dimmer(self, dimmer, *, index=0):
        """Set dimmer value of a light.

        Integer between 0..255
        """
        self.set_values({
            ATTR_LIGHT_DIMMER: dimmer,
        }, index=index)

    def set_hex_color(self, color, *, index=0):
        """Set xy color of the light."""
        self.set_values({
            ATTR_LIGHT_COLOR: color,
        }, index=index)

    def set_xy_color(self, color_x, color_y, *, index=0):
        """Set xy color of the light."""
        self.set_values({
            ATTR_LIGHT_COLOR_X: color_x,
            ATTR_LIGHT_COLOR_Y: color_y
        }, index=index)

    def set_values(self, values, *, index=0):
        """Set values on light control."""
        assert len(self.raw) == 1, \
            'Only devices with 1 light supported'

        self._device.api('put', self._device.path, {
            ATTR_LIGHT_CONTROL: [
                values
            ]
        })

    def __repr__(self):
        return '<LightControl for {} ({} lights)>'.format(self._device.name,
                                                          len(self.raw))


class Light:
    """Represent a light control.

    https://github.com/IPSO-Alliance/pub/blob/master/docs/IPSO-Smart-Objects.pdf
    """

    def __init__(self, device, index):
        self.device = device
        self.index = index

    @property
    def state(self):
        return self.raw.get(ATTR_LIGHT_STATE) == 1

    @property
    def dimmer(self):
        return self.raw.get(ATTR_LIGHT_DIMMER)

    @property
    def hex_color(self):
        return self.raw.get(ATTR_LIGHT_COLOR)

    @property
    def xy_color(self):
        return (self.raw.get(ATTR_LIGHT_COLOR_X),
                self.raw.get(ATTR_LIGHT_COLOR_Y))

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self.device.raw[ATTR_LIGHT_CONTROL][self.index]

    def __repr__(self):
        state = "on" if self.state else "off"
        return "<Light #{} - {}>".format(self.index, state)



"""Represent the gateway."""
from datetime import datetime

from .const import (
    ROOT_DEVICES, ROOT_GROUPS, ROOT_MOODS, PATH_GATEWAY_INFO,
    ATTR_NTP, ATTR_FIRMWARE_VERSION, ATTR_CURRENT_TIME_UNIX,
    ATTR_CURRENT_TIME_ISO8601, ATTR_FIRST_SETUP, ATTR_GATEWAY_ID)
from .device import Device
from .group import Group
from .mood import Mood


class Gateway(object):
    """This class connects to the IKEA Tradfri Gateway"""

    def __init__(self, api):
        self.api = api

    def get_endpoints(self):
        """Return all available endpoints on the gateway."""
        data = self.api('get', ['.well-known', 'core'], parse_json=False)
        return [line.split(';')[0][2:-1] for line in data.split(',')]

    def get_devices(self):
        """Returns the devices linked to the gateway"""
        devices = self.api('get', [ROOT_DEVICES])

        return [self.get_device(dev) for dev in devices]

    def get_device(self, device_id):
        """Return specified device."""
        return Device(self.api, self.api('get', [ROOT_DEVICES, device_id]))

    def get_groups(self):
        """Return the groups linked to the gateway."""
        groups = self.api('get', [ROOT_GROUPS])

        return [self.get_group(group) for group in groups]

    def get_group(self, group_id):
        """Return specified group."""
        return Group(self, self.api('get', [ROOT_GROUPS, group_id]))

    def get_gateway_info(self):
        """Return the gateway info."""
        return GatewayInfo(self.api, self.api('get', PATH_GATEWAY_INFO))

    def get_moods(self):
        """Return moods defined on the gateway."""
        mood_parent = self._get_mood_parent()

        return [self.get_mood(mood, mood_parent=mood_parent) for mood in
                self.api('get', [ROOT_MOODS, mood_parent])]

    def get_mood(self, mood_id, *, mood_parent=None):
        """Return a mood."""
        if mood_parent is None:
            mood_parent = self._get_mood_parent()

        return Mood(
            self.api, self.api('get', [ROOT_MOODS, mood_parent, mood_id]),
            mood_parent)

    def _get_mood_parent(self):
        """Get the parent of all moods."""
        return self.api('get', [ROOT_MOODS])[0]


class GatewayInfo:
    def __init__(self, api, raw):
        self.api = api
        self.raw = raw

    @property
    def id(self):
        """This looks like a value representing an id."""
        return self.raw.get(ATTR_GATEWAY_ID)

    @property
    def ntp_server(self):
        """NTP server in use."""
        return self.raw.get(ATTR_NTP)

    @property
    def firmware_version(self):
        """NTP server in use."""
        return self.raw.get(ATTR_FIRMWARE_VERSION)

    @property
    def current_time(self):
        if ATTR_CURRENT_TIME_UNIX not in self.raw:
            return None
        return datetime.utcfromtimestamp(self.raw[ATTR_CURRENT_TIME_UNIX])

    @property
    def current_time_iso8601(self):
        return self.raw.get(ATTR_CURRENT_TIME_ISO8601)

    @property
    def first_setup(self):
        """This is a guess of the meaning of this value."""
        if ATTR_FIRST_SETUP not in self.raw:
            return None
        return datetime.utcfromtimestamp(self.raw[ATTR_FIRST_SETUP])

    @property
    def path(self):
        return PATH_GATEWAY_INFO

    def set_values(self, values):
        """Helper to set values for mood."""
        self.api('put', self.path, values)

    def update(self):
        self.raw = self.api('get', self.path)

    def __repr__(self):
        return '<GatewayInfo>'



"""Represent a mood on the gateway."""
from datetime import datetime

from .const import (
    ROOT_MOODS, ATTR_ID, ATTR_NAME, ATTR_CREATED_AT)


class Mood:
    def __init__(self, api, raw, parent):
        self.api = api
        self.raw = raw
        self._parent = parent

    @property
    def id(self):
        return self.raw.get(ATTR_ID)

    @property
    def name(self):
        return self.raw.get(ATTR_NAME)

    @property
    def created_at(self):
        if ATTR_CREATED_AT not in self.raw:
            return None
        return datetime.utcfromtimestamp(self.raw[ATTR_CREATED_AT])

    @property
    def path(self):
        return [ROOT_MOODS, self._parent, self.id]

    def set_values(self, values):
        """Helper to set values for mood."""
        self.api('put', self.path, values)

    def update(self):
        self.raw = self.api('get', self.path)

    def __repr__(self):
        return '<Mood {}>'.format(self.name)



"""Coap implementation."""
import json
import logging
import subprocess

from .error import RequestError, ClientError, ServerError, RequestTimeout

_LOGGER = logging.getLogger(__name__)


CLIENT_ERROR_PREFIX = '4.'
SERVER_ERROR_PREFIX = '5.'


def api_factory(host, security_code):
    """Generate a request method."""
    def request(method, path, data=None, *, parse_json=True, timeout=10):
        """Make a request."""

        path = '/'.join(str(v) for v in path)
        command_string = 'coaps://{}:5684/{}'.format(host, path)

        command = [
            'coap-client',
            '-u',
            'Client_identity',
            '-k',
            security_code,
            '-v',
            '0',
            '-m',
            method,
            command_string
        ]

        kwargs = {
            'timeout': timeout,
            'stderr': subprocess.STDOUT,
        }

        if data is not None:
            kwargs['input'] = json.dumps(data).encode('utf-8')
            command.append('-f')
            command.append('-')
            _LOGGER.debug('Executing %s %s %s: %s', host, method, path, data)
        else:
            _LOGGER.debug('Executing %s %s %s', host, method, path)

        try:
            return_value = subprocess.check_output(command, **kwargs)
            out = return_value.strip().decode('utf-8')
        except subprocess.TimeoutExpired:
            raise RequestTimeout() from None
        except subprocess.CalledProcessError as err:
            raise RequestError(
                'Error executing request: %s'.format(err)) from None

        # Return only the last line, where there's JSON
        lines = [line for line in out.split('\n')[1:]
                 if 'decrypt_verify' not in line]

        if not lines:
            return None

        output = lines[0]
        _LOGGER.debug('Received: %s', output)

        if output.startswith(CLIENT_ERROR_PREFIX):
            raise ClientError(output)

        elif output.startswith(SERVER_ERROR_PREFIX):
            raise ServerError(output)

        elif not parse_json:
            return output

        return json.loads(output)

    # This will cause a RequestError to be raised if credentials invalid
    request('get', ['status'])

    return request



ROOT_DEVICES = "15001"
ROOT_GROUPS = "15004"
ROOT_MOODS = "15005"
PATH_GATEWAY_INFO = ["15011", "15012"]

ATTR_APPLICATION_TYPE = "5750"
ATTR_DEVICE_INFO = "3"
ATTR_NAME = "9001"
ATTR_CREATED_AT = "9002"
ATTR_ID = "9003"
ATTR_REACHABLE_STATE = "9019"
ATTR_LAST_SEEN = "9020"
ATTR_LIGHT_CONTROL = "3311"  # array

ATTR_NTP = "9023"
ATTR_FIRMWARE_VERSION = "9029"
ATTR_CURRENT_TIME_UNIX = "9059"
ATTR_CURRENT_TIME_ISO8601 = "9060"
ATTR_FIRST_SETUP = "9069"  # ??? unix epoch value when gateway first setup
ATTR_GATEWAY_ID = "9081"  # ??? id of the gateway

ATTR_LIGHT_STATE = "5850"  # 0 / 1
ATTR_LIGHT_DIMMER = "5851"  # Dimmer, not following spec: 0..255
ATTR_LIGHT_COLOR = "5706"  # string representing a value in some color space
ATTR_LIGHT_COLOR_X = "5709"
ATTR_LIGHT_COLOR_Y = "5710"



"""Implement an API wrapper around Ikea Tradfri."""

from .api import retry_timeout
from .coap_cli import api_factory as cli_api_factory
from .error import (
    PyTradFriError, RequestError, ClientError, ServerError, RequestTimeout)
from .gateway import Gateway

__all__ = ['Gateway', 'cli_api_factory', 'PyTradFriError', 'RequestError',
           'ClientError', 'ServerError', 'RequestTimeout', 'retry_timeout']



from setuptools import setup
setup(
  name = 'pytradfri',
  packages = ['pytradfri'],
  version = '0.5',
  description = 'A python library for communicating with the IKEA Tradfri Gateway',
  author = 'balloob, ggravlingen',
  author_email = 'no@email.com',
  url = 'https://github.com/ggravlingen/pytradfri',
  download_url = 'https://github.com/ggravlingen/pytradfri/archive/v0.5.zip'
)



"""API utilities."""

from .error import RequestTimeout


def retry_timeout(api, retries=3):
    """Retry API call when a timeout occurs."""
    def retry_api(*args, **kwargs):
        """Retrying API."""
        for i in range(1, retries + 1):
            try:
                return api(*args, **kwargs)
            except RequestTimeout:
                if i == retries:
                    raise

    return retry_api
